/* LSDADecoder
Copyright (C) 2019  Pavel Cernohorsky

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.*/

#pragma once

#include <cstdint>
#include <optional>
#include <vector>

/**
 * DWARF Exception Header Encoding
 * http://refspecs.linuxbase.org/LSB_5.0.0/LSB-Core-generic/LSB-Core-generic/dwarfext.html */
using DWARFEHEncoding = uint8_t;
struct DWARFEHEncodingVal
{
  static const uint8_t DW_EH_PE_absptr  = 0x00; // The Value is a literal pointer whose size is determined by the architecture.
  static const uint8_t DW_EH_PE_uleb128 = 0x01; // Unsigned value is encoded using the Little Endian Base 128 (LEB128) as defined by DWARF Debugging Information Format, Version 4.
  static const uint8_t  DW_EH_PE_udata2 = 0x02; // A 2 bytes unsigned value.
  static const uint8_t DW_EH_PE_udata4  = 0x03; // A 4 bytes unsigned value.
  static const uint8_t DW_EH_PE_udata8  = 0x04; // An 8 bytes unsigned value.
  static const uint8_t DW_EH_PE_sleb128 = 0x09; // Signed value is encoded using the Little Endian Base 128 (LEB128) as defined by DWARF Debugging Information Format, Version 4.
  static const uint8_t DW_EH_PE_sdata2  = 0x0A; // A 2 bytes signed value.
  static const uint8_t DW_EH_PE_sdata4  = 0x0B; // A 4 bytes signed value.
  static const uint8_t DW_EH_PE_sdata8  = 0x0C; // An 8 bytes signed value.

  static const uint8_t DW_EH_PE_pcrel   = 0x10; // Value is relative to the current program counter.
  static const uint8_t DW_EH_PE_textrel = 0x20; // Value is relative to the beginning of the .text section.
  static const uint8_t DW_EH_PE_datarel = 0x30; // Value is relative to the beginning of the .got or .eh_frame_hdr section.
  static const uint8_t DW_EH_PE_funcrel = 0x40; // Value is relative to the beginning of the function.
  static const uint8_t DW_EH_PE_aligned = 0x50; // Value is aligned to an address unit sized boundary.

  static const uint8_t DW_EH_PE_omit = 0xFF; // Shall be used to indicate that no value is present.
};

//------------------------------------------------------------------------------

struct LSDAHeader
{
  DWARFEHEncoding m_landingPadBaseEncoding;
  /**
   * If this is omitted, the base comes from calling _Unwind_GetRegionStart, which returns the
   * beginning of the code described by the current FDE. */
  std::optional<int64_t> m_landingPadBase;
  DWARFEHEncoding m_typeTableEntryEncoding;
  /**
   * Type table offeset from this filed. */
  std::optional<size_t> m_typeTableOffset;
  std::optional<size_t> m_typeTableStart;
  DWARFEHEncoding m_callSiteTableEntryEncoding;
  size_t m_callSiteTableSize;
  size_t m_actionTableStart;
};

//------------------------------------------------------------------------------

struct LSDAType
{
  /**
   * Value of what is defined by the m_filter in LSDAAction. */
  int64_t m_typeOffset;
  /**
   * The actual value of where the type is stored (m_typeOffset + where the offset is stored). */
  int64_t m_type;
  /**
   * Indication that the m_type should use indirect addressing - the value of m_type points to
   * address which holds the actual address of the type. */
  bool m_typeIsIndirect;
};

//------------------------------------------------------------------------------

struct LSDAAction
{
  /**
   * Just for orientation - absolute offset while beginning to parse the value. */
  size_t m_address;
  /**
   * Type filter, 0 means cleanup only. Positive value is index into the types table. This is a
   * negative index: the value 1 means the entry preceding the types table base, 2 means the entry
   * before that, etc.*/
  int64_t m_filter;
  /**
   * Offset to the next action in the current action chain. The offset is relative to the current
   * element, which means that value of 1 (which has 1 byte in sleb128 encoding) points to action
   * precisely following this action. */
  int64_t m_disp;
  /**
   * Details of the type stored in the action. */
  std::optional<LSDAType> m_type;
};

//------------------------------------------------------------------------------

struct LSDACallSite
{
  /**
   * Byte offset from the m_landingPadBase in the header. */
  int64_t m_ipStart;
  /**
   * Length in bytes of the instruction pointer range. */
  int64_t m_ipLength;
  /**
   * Byte offset from the m_landingPadBase in the header. */
  int64_t m_landingPad;
  /**
   * 0 for no action, otherwise 1 + byte offset into the action table */
  uint64_t m_action;
  /**
   * Chain of actions parsed out based on the m_action. */
  std::vector<LSDAAction> m_actions;
};

//------------------------------------------------------------------------------

struct LSDA
{
  LSDAHeader m_header;
  std::vector<LSDACallSite> m_callSiteTable;
};
