/* LSDADecoder
Copyright (C) 2019  Pavel Cernohorsky

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.*/

#include "LSDAParser.h"

#include "DWARFPrinter.h"

#include <cstring>
#include <stdexcept>
#include <tuple>

namespace
{

void ensureRemainig(size_t offset, size_t maxOffset, size_t needAtLeast)
{
  if (offset + needAtLeast - 1 > maxOffset)
  {
    throw std::runtime_error("Error while reading data, offset " + std::to_string(offset) +
                             ", need at least " + std::to_string(needAtLeast) + ", maxOffset " +
                             std::to_string(maxOffset) + ".");
  }
}

//------------------------------------------------------------------------------

DWARFEHEncoding readDWARFEHEncoding(const uint8_t* startPtr, size_t& offset, const size_t maxOffset)
{
  ensureRemainig(offset, maxOffset, 1);
  uint8_t result = startPtr[offset];
  ++offset;
  return static_cast<DWARFEHEncoding>(result);
}

//------------------------------------------------------------------------------

uint64_t readDWARFUleb128(const uint8_t* startPtr, size_t& offset, const size_t maxOffset)
{
  uint64_t result = 0;
  uint64_t shift = 0;
  uint64_t byteVal;
  do
  {
    ensureRemainig(offset, maxOffset, 1);
    byteVal = startPtr[offset];
    ++offset;
    if (shift >= (sizeof(byteVal) * 8))
    {
      throw std::runtime_error("Overflow while reading ULEB128, current offset " +
                               std::to_string(offset) + ", current shift " +
                               std::to_string(shift) + ".");
    }
    result |= (byteVal & 0x7F) << shift;
    shift += 7;
  } while (byteVal & 0x80);
  return result;
}

//------------------------------------------------------------------------------

int64_t readDWARFSleb128(const uint8_t* startPtr, size_t& offset, const size_t maxOffset)
{
  int64_t result = 0;
  int64_t shift = 0;
  int64_t byteVal;
  bool overflow = false;
  do
  {
    ensureRemainig(offset, maxOffset, 1);
    byteVal = startPtr[offset];
    ++offset;
    if (shift >= static_cast<int64_t>(sizeof(byteVal) * 8))
    {
      overflow = true;
      break;
    }
    result |= (byteVal & 0x7F) << shift;
    shift += 7;
  } while (byteVal & 0x80);

  if (shift >= static_cast<int64_t>(sizeof(byteVal) * 8))
  {
    overflow = true;
  }
  if (overflow)
  {
    throw std::runtime_error("Overflow while reading SLEB128, current offset " +
                             std::to_string(offset) + ", current shift " +
                             std::to_string(shift) + ".");
  }
  if (byteVal & 0x40)
  {
    //result |= (~0 << shift);
    uint64_t signExtension = 0xFFFFFFFFFFFFFFFF << shift;
    result |= static_cast<int64_t>(signExtension);
  }
  return result;
}

//------------------------------------------------------------------------------

int32_t readDWARFSdata4(const uint8_t* startPtr, size_t& offset, const size_t maxOffset)
{
  int32_t result;
  ensureRemainig(offset, maxOffset, 4);
  memcpy(&result, startPtr + offset, 4);
  offset += 4;
  return result;
}

//------------------------------------------------------------------------------

std::optional<int64_t> readDwarfEncodedValue(DWARFEHEncoding encoding, const uint8_t* startPtr,
                                             size_t& offset, const size_t maxOffset)
{
  if (encoding == DWARFEHEncodingVal::DW_EH_PE_omit)
  {
    return {};
  }
  if ((encoding & DWARFEHEncodingVal::DW_EH_PE_uleb128) == DWARFEHEncodingVal::DW_EH_PE_uleb128)
  {
    return readDWARFUleb128(startPtr, offset, maxOffset);
  }
  if ((encoding & DWARFEHEncodingVal::DW_EH_PE_sleb128) == DWARFEHEncodingVal::DW_EH_PE_sleb128)
  {
    return readDWARFSleb128(startPtr, offset, maxOffset);
  }
  throw std::logic_error("Unsupported DWARF encoding \"" + DWARFPrinter::encodingToStr(encoding) +
                         "\", offset " + std::to_string(offset));
}

//------------------------------------------------------------------------------

LSDAHeader parseHeader(const uint8_t* startPtr, size_t& offset, const size_t maxOffset)
{
  LSDAHeader result;

  result.m_landingPadBaseEncoding = readDWARFEHEncoding(startPtr, offset, maxOffset);
  result.m_landingPadBase = readDwarfEncodedValue(result.m_landingPadBaseEncoding, startPtr,
                                                  offset, maxOffset);
  result.m_typeTableEntryEncoding = readDWARFEHEncoding(startPtr, offset, maxOffset);
  if (result.m_typeTableEntryEncoding != DWARFEHEncodingVal::DW_EH_PE_omit)
  {
    result.m_typeTableOffset = readDWARFUleb128(startPtr, offset, maxOffset);
    // make sure that there is no type base, because we ignore that one for now
    DWARFEHEncoding maskedEncoding = result.m_typeTableEntryEncoding & 0x70;
    if (maskedEncoding != DWARFEHEncodingVal::DW_EH_PE_absptr &&
        maskedEncoding != DWARFEHEncodingVal::DW_EH_PE_pcrel &&
        maskedEncoding != DWARFEHEncodingVal::DW_EH_PE_aligned)
    {
      throw std::logic_error("Type base is currently not supported, see unwind-pe.h:101 "
                             "(base_of_encoded_value).");
    }
  }
  else
  {
    result.m_typeTableOffset = std::nullopt;
  }

  if (result.m_typeTableOffset)
  {
    result.m_typeTableStart = offset + static_cast<size_t>(*result.m_typeTableOffset);
  }
  else
  {
    result.m_typeTableStart = 0;
  }
  result.m_callSiteTableEntryEncoding = readDWARFEHEncoding(startPtr, offset, maxOffset);
  result.m_callSiteTableSize = readDWARFUleb128(startPtr, offset, maxOffset);
  result.m_actionTableStart = offset + result.m_callSiteTableSize;
  return result;
}

//------------------------------------------------------------------------------

LSDACallSite parseCallSiteTableEntry(const uint8_t* startPtr, size_t& offset,
                                     const size_t maxOffset, DWARFEHEncoding encoding)
{
  LSDACallSite result;
  result.m_ipStart = readDwarfEncodedValue(encoding, startPtr, offset, maxOffset).value();
  result.m_ipLength = readDwarfEncodedValue(encoding, startPtr, offset, maxOffset).value();
  result.m_landingPad = readDwarfEncodedValue(encoding, startPtr, offset, maxOffset).value();
  result.m_action = readDWARFUleb128(startPtr, offset, maxOffset);
  return result;
}

//------------------------------------------------------------------------------

std::vector<LSDACallSite> parseCallSiteTable(const uint8_t* startPtr, size_t& offset,
                                             const size_t maxOffset, DWARFEHEncoding encoding,
                                             size_t size)
{
  std::vector<LSDACallSite> result;
  while (size)
  {
    size_t startOffset = offset;
    result.push_back(parseCallSiteTableEntry(startPtr, offset, maxOffset, encoding));
    size_t parsedSize = offset - startOffset;
    if (size < parsedSize)
    {
      throw std::runtime_error("Call site table alignment problem, remaining size " +
                               std::to_string(size) + " but parsed " + std::to_string(parsedSize));
    }
    size -= parsedSize;
  }
  return result;
}

//------------------------------------------------------------------------------

std::vector<LSDAAction> parseActionChain(const uint8_t* startPtr, size_t offset,
                                         const size_t maxOffset)
{
  std::vector<LSDAAction> result;
  while (true)
  {
    LSDAAction action;
    action.m_address = offset;
    action.m_filter = readDWARFSleb128(startPtr, offset, maxOffset);
    size_t currentOffset = offset;
    action.m_disp = readDWARFSleb128(startPtr, offset, maxOffset);
    result.push_back(action);
    if (!action.m_disp)
    {
      break;
    }
    else
    {
      offset = static_cast<size_t>(static_cast<int64_t>(currentOffset) + action.m_disp);
    }
  }
  return result;
}

//------------------------------------------------------------------------------

/**
 * Fills in action chains into the given call site table. */
void parseActions(const uint8_t* startPtr, const size_t maxOffset, size_t actionTableStart,
                  std::vector<LSDACallSite>& callSiteTable)
{
  for (auto& callSiteTableEntry: callSiteTable)
  {
    if (callSiteTableEntry.m_action)
    {
      size_t actionChainStart = actionTableStart + callSiteTableEntry.m_action - 1;
      callSiteTableEntry.m_actions = parseActionChain(startPtr, actionChainStart, maxOffset);
    }
  }
}

//------------------------------------------------------------------------------

int64_t sizeOfEncodedType(DWARFEHEncoding encoding)
{
  if (encoding == DWARFEHEncodingVal::DW_EH_PE_omit)
  {
    return 0;
  }
  switch (encoding & 0x07)
  {
    case DWARFEHEncodingVal::DW_EH_PE_absptr:
      return sizeof(void*);
    case DWARFEHEncodingVal::DW_EH_PE_udata2:
      return 2;
    case DWARFEHEncodingVal::DW_EH_PE_udata4:
      return 4;
    case DWARFEHEncodingVal::DW_EH_PE_udata8:
      return 8;
  }
  throw std::runtime_error("Unsupported type encoding (size of value) " +
                           std::to_string(static_cast<uint16_t>(encoding)) + '.');
}

//------------------------------------------------------------------------------

LSDAType parseType(const uint8_t* startPtr, const size_t maxOffset, const size_t typeTableStart,
                   DWARFEHEncoding encoding, int64_t filter)
{
  const int64_t tableOffset = filter * sizeOfEncodedType(encoding);
  if (static_cast<size_t>(tableOffset) > typeTableStart)
  {
    throw std::runtime_error("Type reading failed, filter higher than table start.");
  }
  const size_t offset = typeTableStart - static_cast<size_t>(tableOffset);
  int64_t resultOffset = 0;
  int64_t result = 0;
  bool resultIsIndirect = false;
  if (encoding == DWARFEHEncodingVal::DW_EH_PE_aligned)
  {
    throw std::logic_error("Unsupported in parseType 1");
  }
  else
  {
    DWARFEHEncoding maskedEncoding = encoding & 0x0F;
    switch (maskedEncoding)
    {
      case DWARFEHEncodingVal::DW_EH_PE_absptr:
        throw std::logic_error("Unsupported in parseType 2");
        break;

      case DWARFEHEncodingVal::DW_EH_PE_uleb128:
        throw std::logic_error("Unsupported in parseType 3");
        break;

      case DWARFEHEncodingVal::DW_EH_PE_sleb128:
        throw std::logic_error("Unsupported in parseType 4");
        break;

      case DWARFEHEncodingVal::DW_EH_PE_udata2:
        throw std::logic_error("Unsupported in parseType 5");
        break;

      case DWARFEHEncodingVal::DW_EH_PE_udata4:
        throw std::logic_error("Unsupported in parseType 6");
        break;

      case DWARFEHEncodingVal::DW_EH_PE_udata8:
        throw std::logic_error("Unsupported in parseType 7");
        break;

      case DWARFEHEncodingVal::DW_EH_PE_sdata2:
        throw std::logic_error("Unsupported in parseType 8");
        break;

      case DWARFEHEncodingVal::DW_EH_PE_sdata4:
      {
        auto dummy = offset;
        resultOffset = readDWARFSdata4(startPtr, dummy, maxOffset);
        break;
      }

      case DWARFEHEncodingVal::DW_EH_PE_sdata8:
        throw std::logic_error("Unsupported in parseType 10");
        break;

      default:
        throw std::logic_error("Unsupported type encoding");
    }

    if (resultOffset != 0)
    {
      if ((encoding & 0x70) == DWARFEHEncodingVal::DW_EH_PE_pcrel)
      {
        result = static_cast<int64_t>(offset) + resultOffset;
      }
      else
      {
        // TODO add support for type base
      }
      if (encoding & DWARFEHEncodingVal::DW_EH_PE_pcrel)
      {
        resultIsIndirect = true;
      }
    }
  }

  return {resultOffset, result, resultIsIndirect};
}

//------------------------------------------------------------------------------

/**
 * Fills in types into the action chains of the given call site table. */
void parseTypes(const uint8_t* startPtr, const size_t maxOffset, const size_t typeTableStart,
                DWARFEHEncoding encoding, std::vector<LSDACallSite>& callSiteTable)
{
  for (auto& callSiteTableEntry: callSiteTable)
  {
    for (auto& action: callSiteTableEntry.m_actions)
    {
      if (action.m_filter > 0)
      {
        action.m_type = parseType(startPtr, maxOffset, typeTableStart, encoding, action.m_filter);
      }
      else
      {
        action.m_type = std::nullopt;
      }
    }
  }
}

} // namespace

//------------------------------------------------------------------------------

LSDA LSDAParser::parse(const std::vector<uint8_t>& elfData, uint64_t lsdaOffset)
{
  if (!elfData.size())
  {
    throw std::runtime_error("No LSDA data for parsing.");
  }
  const uint8_t* dataPtr = elfData.data();
  size_t offset = lsdaOffset;
  size_t maxOffset = elfData.size() - 1;

  LSDA result;
  result.m_header = parseHeader(dataPtr, offset, maxOffset);
  result.m_callSiteTable = parseCallSiteTable(dataPtr, offset, maxOffset,
                                              result.m_header.m_callSiteTableEntryEncoding,
                                              result.m_header.m_callSiteTableSize);
  if (offset != result.m_header.m_actionTableStart)
  {
    throw std::runtime_error("Unexpected action table offset " + std::to_string(offset) +
                             ", expected " + std::to_string(result.m_header.m_actionTableStart) +
                             '.');
  }
  parseActions(dataPtr, maxOffset, result.m_header.m_actionTableStart, result.m_callSiteTable);
  if (result.m_header.m_typeTableStart)
  {
    parseTypes(dataPtr, maxOffset, *result.m_header.m_typeTableStart,
               result.m_header.m_typeTableEntryEncoding, result.m_callSiteTable);
  }
  return result;
}
