/* LSDADecoder
Copyright (C) 2019  Pavel Cernohorsky

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.*/

#include "LSDAParser.h"
#include "LSDAPrinter.h"

#include <filesystem>
#include <fstream>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <vector>

//------------------------------------------------------------------------------

std::vector<uint8_t> readFile(const std::filesystem::path& path)
{
  std::ifstream file(path, std::ios::binary | std::ios::ate);
  if (!file)
  {
    throw std::runtime_error("Error while opening file: " + path.string());
  }
  auto size = file.tellg();
  file.seekg(0, std::ios::beg);
  std::vector<uint8_t> result(static_cast<size_t>(size));
  if (!file.read(reinterpret_cast<char*>(result.data()), size))
  {
    throw std::runtime_error("Error while reading data from file: " + path.string());
  }
  return result;
}

//------------------------------------------------------------------------------

uint64_t readHex(const char* value)
{
  std::istringstream input(value);
  uint64_t result;
  input >> std::hex >> result;
  return result;
}

//------------------------------------------------------------------------------

int main(int argc, const char* argv[])
{
  if (argc != 3)
  {
    std::cout << "Utility for decoding of LSDA from the .gcc_except_table ELF section.\n\n";
    std::cout << "Usage:\n  " << argv[0] << " <elf_file_name> <relevant_lsda_offset_hex>\n\n";
    std::cout << "To get the relevant offset, consult with your debugger. Just as an example, the "
                 "following commands will print the first LSDA record stored in the "
                 ".gcc_except_table and help you see the catch types there:\n"
                 "  $ readelf -S TestApp/TestApp | grep gcc_except_table\n"
                 "    [19] .gcc_except_table PROGBITS         0000000000002228  00002228\n"
                 "  $ ./LSDADecoder TestApp/TestApp 0x2228\n"
                 "  $ ./LSDADecoder TestApp/TestApp 0x2228 | grep 'type '\n"
                 "      type 0x0000000000004010\n"
                 "      type 0x0000000000004018\n"
                 "  $ readelf --symbols -W TestApp/TestApp | grep '4010\\|4018'\n"
                 "    67: 0000000000004018     8 OBJECT  WEAK   HIDDEN    25 "
                 "DW.ref._ZTISt9exception\n"
                 "    73: 0000000000004010     8 OBJECT  WEAK   HIDDEN    25 "
                 "DW.ref._ZTI19FileAccessException";
    std::cout << std::endl;
    return 1;
  }
  try
  {
    const std::vector<uint8_t> fileData = readFile(argv[1]);
    uint64_t lsdaOffset = readHex(argv[2]);
    LSDA lsda = LSDAParser::parse(fileData, lsdaOffset);
    LSDAPrinter::print(lsda);
  }
  catch (const std::exception& ex)
  {
    std::cout << "Exception while running: " << ex.what() << std::endl;
    return 2;
  }
  return 0;
}
